# coding: utf-8
# Project Configuration Information

# project info
__version__ = "0.0.20"
__index_page__ = "https://github.com/CzaOrz/loggus"
__doc_page__ = "https://github.com/CzaOrz/loggus/README.md"
# author info
__author__ = "https://github.com/CzaOrz"
__email__ = "972542655@qq.com"
__phone__ = "*** **** ****"



# coding: utf-8
import logging


def IsLevel(level):
    return level in GetAllLevels()


def GetAllLevels():
    return [DEBUG, INFO, WARNING, ERROR, PANIC]


class Level:

    def __init__(self, string: str, integer: int, color: str):
        self.describe = string
        self.value = integer
        self.color = color.format(string)
        self.colorTmp = color

    def toColor(self, msg: str):
        return self.colorTmp.format(msg)

    def __eq__(self, other):
        return self.value == other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __hash__(self):
        return self.value

    def __str__(self):
        return self.describe

    __repr__ = __str__


DEBUG = Level("debug", logging.DEBUG, "\033[1;37m{}\033[0m")
INFO = Level("info", logging.INFO, "\033[1;32m{}\033[0m")
WARNING = Level("warning", logging.WARNING, "\033[1;33m{}\033[0m")
ERROR = Level("error", logging.ERROR, "\033[1;31m{}\033[0m")
PANIC = Level("panic", logging.FATAL, "\033[1;36m{}\033[0m")



# coding: utf-8
import os
import re
import time

from loggus.level import GetAllLevels

ST_MODE = 0
ST_INO = 1
ST_DEV = 2
ST_NLINK = 3
ST_UID = 4
ST_GID = 5
ST_SIZE = 6
ST_ATIME = 7
ST_MTIME = 8
ST_CTIME = 9
MIDNIGHT = 24 * 60 * 60


def IsIHook(hook) -> bool:
    return isinstance(hook, IHook)


def SafeMoveFile(source, destination):
    if not os.path.exists(source):
        return
    if os.path.exists(destination):
        os.remove(destination)
    os.rename(source, destination)


class IHook:

    def GetLevels(self) -> list:
        raise NotImplementedError

    def Fire(self, entry, level, msg, output) -> None:
        raise NotImplementedError


class FileHook(IHook):
    stream = None

    def __init__(self, filename, mode="a", encoding="utf-8"):
        self.filename = os.path.abspath(filename)
        self.mode = mode
        self.encoding = encoding
        self.stream = open(filename, mode=mode, encoding=encoding)

    def GetLevels(self):
        return GetAllLevels()

    def Fire(self, entry, level, msg, output) -> None:
        self.stream.write(output)
        self.stream.flush()

    def __del__(self):
        if hasattr(self.stream, "close"):
            self.stream.close()


class RotatingFileHook(FileHook):

    def __init__(self, filename, mode="a", encoding="utf-8", maxBytes=1024 * 64, backupCount=3):
        super(RotatingFileHook, self).__init__(filename, mode, encoding)
        self.maxBytes = maxBytes
        self.backupCount = backupCount

    def Fire(self, entry, level, msg, output) -> None:
        if self.NeedRollover(output):
            self.DoRollover()
        self.stream.write(output)
        self.stream.flush()

    def NeedRollover(self, output) -> bool:
        self.stream.seek(0, 2)
        if self.stream.tell() + len(output) >= self.maxBytes:
            return True
        return False

    def DoRollover(self):
        if hasattr(self.stream, "close"):
            self.stream.close()
            self.stream = None
        for index in range(self.backupCount - 1, 0, -1):
            SafeMoveFile(f"{self.filename}.{index}", f"{self.filename}.{index + 1}")
        SafeMoveFile(self.filename, f"{self.filename}.1")
        self.stream = open(self.filename, mode=self.mode, encoding=self.encoding)


class TimedRotatingFileHook(RotatingFileHook):

    def __init__(self, filename, when="h", interval=1, mode="a", encoding="utf-8", maxBytes=1024 * 64, backupCount=3):
        super(TimedRotatingFileHook, self).__init__(filename, mode, encoding, maxBytes, backupCount)
        self.when = when.upper()
        self.interval = interval

        if self.when == 'S':
            self.interval = 1  # one second
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when == 'M':
            self.interval = 60  # one minute
            self.suffix = "%Y-%m-%d_%H-%M"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}(\.\w+)?$"
        elif self.when == 'H':
            self.interval = 60 * 60  # one hour
            self.suffix = "%Y-%m-%d_%H"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}(\.\w+)?$"
        elif self.when == 'D' or self.when == 'MIDNIGHT':
            self.interval = 60 * 60 * 24  # one day
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        else:
            raise ValueError("Invalid rollover interval specified: %s" % self.when)
        self.extMatch = re.compile(self.extMatch, re.ASCII)
        self.interval = self.interval * interval  # multiply by units requested
        if os.path.exists(self.filename):
            current = os.stat(self.filename)[ST_MTIME]
        else:
            current = int(time.time())
        self.nextRolloverAt = self.NextRollover(current)
        # TODO: DoRollover Exists Some Bugs.
        raise Exception("TimedRotatingFileHook.DoRollover Exists Some Bugs.")

    def NextRollover(self, current):
        return current + self.interval

    def NeedRollover(self, output) -> bool:
        if int(time.time()) >= self.nextRolloverAt:
            return True
        return False

    def DoRollover(self):
        if hasattr(self.stream, "close"):
            self.stream.close()
            self.stream = None

        nextRolloverAt = self.nextRolloverAt

        for index in range(self.backupCount - 1, 0, -1):
            previousFileName = f"{self.filename}.{time.strftime(self.suffix, time.localtime(nextRolloverAt - self.interval * index))}"
            nextFileName = f"{self.filename}.{time.strftime(self.suffix, time.localtime(nextRolloverAt - self.interval * (index - 1)))}"
            SafeMoveFile(previousFileName, nextFileName)
            nextRolloverAt -= self.interval
        SafeMoveFile(self.filename, f"{self.filename}.{time.strftime(self.suffix, time.localtime(nextRolloverAt))}")
        self.stream = open(self.filename, mode=self.mode, encoding=self.encoding)
        self.nextRolloverAt = self.NextRollover(self.nextRolloverAt)



# coding: utf-8
import re

from datetime import datetime

regex = re.compile("^[a-zA-Z0-9]*$")


def IsIField(*fields) -> bool:
    for field in fields:
        if field.__class__ is not IFieldMetaClass:
            return False
    return True


class IFieldMetaClass(type):

    def __new__(cls, name: str, bases: tuple, attrs: dict):
        if bases:
            if IField not in bases:
                raise Exception(f"please ensure `{name}` implemented the interface of `loggus.interfaces.IField`")
            if attrs.get("Key", None) is None:
                raise Exception(f"please ensure `{name}` implemented the attr of `Key`")
            if "ResolveIn" not in attrs:
                raise Exception(f"please ensure `{name}` implemented the function of `ResolveIn`")
            if "ResolveOut" not in attrs:
                raise Exception(f"please ensure `{name}` implemented the function of `ResolveOut`")
        return type.__new__(cls, name, bases, attrs)


class IField(metaclass=IFieldMetaClass):
    Key = None

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        raise NotImplementedError

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        raise NotImplementedError


class FieldKeyTime(IField):
    Key = "time"

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        entry.fields[cls.Key] = f"{datetime.now()}"

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        entry.fields.pop(cls.Key, None)
        return f"{cls.Key}=\"{datetime.now()}\""


class FieldKeyLevel(IField):
    Key = "level"

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        entry.fields[cls.Key] = level.describe

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        entry.fields.pop(cls.Key, None)
        return f"{cls.Key}={level.color if entry.logger.colorSwitch else level.describe}"


class FieldKeyMsg(IField):
    Key = "msg"

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        entry.fields[cls.Key] = msg

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        entry.fields.pop(cls.Key, None)
        if regex.match(msg):
            return f"{cls.Key}={msg}"
        else:
            return f"{cls.Key}=\"{msg}\""


class FieldKeyLineNo(IField):
    Key = "lineNo"

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        entry.fields[cls.Key] = entry.frameLineNo

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        return f"{cls.Key}={entry.frameLineNo}"


class FieldKeyFunc(IField):
    Key = "funcName"

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        entry.fields[cls.Key] = entry.frameFuncName

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        return f"{cls.Key}={entry.frameFuncName}"


class FieldKeyFile(IField):
    Key = "filePath"

    @classmethod
    def ResolveIn(cls, entry, level, msg):
        entry.fields[cls.Key] = entry.frameFilePath

    @classmethod
    def ResolveOut(cls, entry, level, msg):
        return f"{cls.Key}={entry.frameFilePath}"



# coding: utf-8
import sys
import traceback
import contextlib

from copy import deepcopy
from _io import TextIOWrapper
from typing import List, Dict

from loggus.level import *
from loggus.hooks import *
from loggus.fields import *
from loggus.formatter import *

# fix messy code when output color in win32.
if sys.platform == "win32":
    import colorama

    colorama.init(autoreset=True)

if hasattr(sys, '_getframe'):
    currentframe = lambda: sys._getframe(3)
else:
    def currentframe():
        try:
            raise Exception
        except Exception:
            return sys.exc_info()[2].tb_frame.f_back


def FindCaller():
    f = currentframe()
    if f is not None:
        f = f.f_back
    rv = "(unknown file)", "(unknown line)", "(unknown function)"
    while hasattr(f, "f_code"):
        co = f.f_code
        filename = os.path.normcase(co.co_filename)
        if filename == _srcfile:
            f = f.f_back
            continue
        rv = (co.co_filename, f.f_lineno, co.co_name)
        break
    return rv


_srcfile = os.path.normcase(FindCaller.__code__.co_filename)


class Logger:
    stream: TextIOWrapper = sys.stdout
    formatter: IFormatter = TextFormatter
    fieldKeys: List[IField] = [FieldKeyTime, FieldKeyLevel, FieldKeyMsg]
    needFrame: bool = False
    baseLevel: Level = INFO
    colorSwitch: bool = True
    hooks: Dict[Level, List[IHook]] = {
        DEBUG: [],
        INFO: [],
        WARNING: [],
        ERROR: [],
        PANIC: [],
    }

    def AddHooks(self, *hooks: List[IHook]) -> None:
        for hook in hooks:  # type: IHook
            if not IsIHook(hook):
                raise Exception(f"undefined Hook[{hook}]!")
            for level in hook.GetLevels():
                if not IsLevel(level):
                    raise Exception(f"undefined Level[{level}]!")
                self.hooks[level].append(hook)

    def FireHooks(self, entry, level, msg, output) -> None:
        for hook in self.hooks[level]:  # type: IHook
            hook.Fire(entry, level, msg, output)

    def IsLevelEnabled(self, level: Level) -> bool:
        return level >= self.baseLevel

    def SetLevel(self, level: Level) -> None:
        if not IsLevel(level):
            raise Exception(f"undefined Level[{level}]!")
        self.baseLevel = level

    def SetFormatter(self, formatter: IFormatter):
        if not IsIFormatter(formatter):
            raise Exception(f"undefined Formatter[{formatter}]")
        if formatter is JsonFormatter:
            self.colorSwitch = False
        self.formatter = formatter

    def SetFieldKeys(self, *fieldKeys):
        fields = []
        for fieldKey in fieldKeys:
            if not IsIField(fieldKey):
                raise Exception(f"undefined Field[{fieldKey}]!")
            fields.append(fieldKey)
        if fields:
            self.fieldKeys = fields
        if {FieldKeyFunc, FieldKeyLineNo, FieldKeyFile} & {*self.fieldKeys}:
            self.needFrame = True
        else:
            self.needFrame = False

    def OpenFieldKeyFunc(self):
        if FieldKeyFunc not in self.fieldKeys:
            self.fieldKeys.append(FieldKeyFunc)
        self.needFrame = True

    def OpenFieldKeyLineNo(self):
        if FieldKeyLineNo not in self.fieldKeys:
            self.fieldKeys.append(FieldKeyLineNo)
        self.needFrame = True

    def OpenFieldKeyFile(self):
        if FieldKeyFile not in self.fieldKeys:
            self.fieldKeys.append(FieldKeyFile)
        self.needFrame = True

    def Write(self, output: str) -> None:
        self.stream.write(output)
        self.stream.flush()

    def NewEntry(self):
        return NewEntry(self)

    def withField(self, key, value, colorLevel=None):
        entry = self.NewEntry()
        return entry.withField(key, value, colorLevel)

    def withFields(self, fields: dict):
        entry = self.NewEntry()
        return entry.withFields(fields)

    def withFieldTrace(self):
        entry = self.NewEntry()
        return entry.withFieldTrace()

    def withTraceback(self, callback=None, *args, **kwargs):
        entry = self.NewEntry()
        return entry.withTraceback(callback, *args, **kwargs)

    def debug(self, *args):
        entry = self.NewEntry()
        entry.debug(*args)

    def info(self, *args):
        entry = self.NewEntry()
        entry.info(*args)

    def warning(self, *args):
        entry = self.NewEntry()
        entry.warning(*args)

    def error(self, *args):
        entry = self.NewEntry()
        entry.error(*args)

    def panic(self, *args):
        entry = self.NewEntry()
        entry.panic(*args)


def NewLogger():
    return Logger()


def SetLevel(level: Level) -> None:
    _logger.SetLevel(level)


def SetFormatter(formatter: TextFormatter or JsonFormatter):
    _logger.SetFormatter(formatter)


def OpenFieldKeyFunc():
    _logger.OpenFieldKeyFunc()


def OpenFieldKeyLineNo():
    _logger.OpenFieldKeyLineNo()


def OpenFieldKeyFile():
    _logger.OpenFieldKeyFile()


def SetFieldKeys(*fieldKeys):
    _logger.SetFieldKeys(*fieldKeys)


def CloseColor():
    _logger.colorSwitch = False


def OpenColor():
    _logger.colorSwitch = True


def AddHook(*hooks):
    _logger.AddHooks(*hooks)


_logger = NewLogger()


class Entry:

    def __init__(self, logger: Logger):
        self.logger = logger
        self.fields = {}
        self.frameFuncName = None
        self.frameLineNo = None
        self.frameFilePath = None

    def withField(self, key, value, colorLevel: Level = None):
        if colorLevel and self.logger.colorSwitch:
            value = colorLevel.toColor(value)
        return self.withFields({key: value})

    def withFields(self, fields: dict):
        if not isinstance(fields, dict):
            raise Exception(f"unsupported {type(fields)}")
        entry = NewEntry(self.logger, self.fields)
        entry.fields.update(fields)
        return entry

    def withFieldTrace(self):
        return self.withField("traceback", traceback.format_exc().strip(), ERROR)

    @contextlib.contextmanager
    def withTraceback(self, callback=None, *args, **kwargs):
        try:
            yield
        except Exception as e:
            if callback:
                try:
                    callback(e, *args, **kwargs)
                except:
                    pass
            self.withFieldTrace().error("an error occurred.")

    def log(self, level: Level, msg: str):
        entry = NewEntry(self.logger, self.fields)
        if entry.logger.needFrame:
            entry.frameFilePath, entry.frameLineNo, entry.frameFuncName, = FindCaller()
        output = entry.logger.formatter.Format(entry, level, msg)
        entry.logger.FireHooks(entry, level, msg, output)
        entry.logger.Write(output)

    def Log(self, level: Level, msg: str):
        if self.logger.IsLevelEnabled(level):
            self.log(level, msg)
            if level >= PANIC:
                sys.exit(PANIC.value)

    def debug(self, *args):
        self.Log(DEBUG, " ".join([f"{arg}" for arg in args]))

    def info(self, *args):
        self.Log(INFO, " ".join([f"{arg}" for arg in args]))

    def warning(self, *args):
        self.Log(WARNING, " ".join([f"{arg}" for arg in args]))

    def error(self, *args):
        self.Log(ERROR, " ".join([f"{arg}" for arg in args]))

    def panic(self, *args):
        self.Log(PANIC, " ".join([f"{arg}" for arg in args]))


def NewEntry(logger: Logger = _logger, fields=None):
    entry = Entry(logger)
    if fields:
        try:
            fields = deepcopy(fields)
        except:
            fields = {f"{key}": f"{value}" for key, value in fields.items()}
        entry.fields = fields
    return entry


def withField(key, value, color: str = None) -> Entry:
    return NewEntry().withField(key, value, color)


def withFields(fields: dict) -> Entry:
    return NewEntry().withFields(fields)


def withFieldTrace():
    return NewEntry().withFieldTrace()


def withTraceback(callback=None, *args, **kwargs):
    return NewEntry().withTraceback(callback, *args, **kwargs)


def debug(*args) -> None:
    entry = NewEntry()
    entry.debug(*args)


def info(*args) -> None:
    entry = NewEntry()
    entry.info(*args)


def warning(*args) -> None:
    entry = NewEntry()
    entry.warning(*args)


def error(*args) -> None:
    entry = NewEntry()
    entry.error(*args)


def panic(*args) -> None:
    entry = NewEntry()
    entry.panic(*args)



# coding: utf-8
import re
import json

from datetime import datetime
from loggus.level import Level

regex = re.compile("^[a-zA-Z0-9]*$")


def IsIFormatter(formatter) -> bool:
    return formatter.__class__ is IFormatterMetaClass


class IFormatterMetaClass(type):

    def __new__(cls, name: str, bases: tuple, attrs: dict):
        if bases:
            if IFormatter not in bases:
                raise Exception(f"please ensure `{name}` implemented the interface of `loggus.interfaces.IFormatter`")
            if "Format" not in attrs:
                raise Exception(f"please ensure `{name}` implemented the function of `Format`")
        return type.__new__(cls, name, bases, attrs)


class IFormatter(metaclass=IFormatterMetaClass):

    def Format(self, entry, level, msg) -> str:
        raise NotImplementedError


#
#  Text Formatter
#


class TextFormatter(IFormatter):

    @classmethod
    def Format(cls, entry, level: Level, msg: str) -> str:
        output = " ".join([fieldKey.ResolveOut(entry, level, msg) for fieldKey in entry.logger.fieldKeys])
        for key, value in entry.fields.items():
            if isinstance(value, str):
                if regex.match(value):
                    output += f" {key}={value}"
                else:
                    output += f" {key}=\"{value}\""
            elif isinstance(value, int):
                output += f" {key}={value}"
            else:
                output += f" {key}=\"{value}\""
        return f"{output}\n"


#
#  Json Formatter
#


class PrettyEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, datetime):
            return str(obj)
        return super().default(obj)


# json encoder: forced every obj to str.
class ForcedEncoder(json.JSONEncoder):

    def default(self, obj) -> str:
        return f"{obj}"


class JsonFormatter(IFormatter):

    @classmethod
    def Format(cls, entry, level, msg) -> str:
        for fieldKey in entry.logger.fieldKeys:
            fieldKey.ResolveIn(entry, level, msg)
        try:
            output = json.dumps(entry.fields, ensure_ascii=False, cls=PrettyEncoder)
        except:
            output = json.dumps(entry.fields, ensure_ascii=False, cls=ForcedEncoder)
        return f"{output}\n"



import os
import re
import codecs

from setuptools import setup, find_packages

"""pip install twine
1、python setup.py check
2、python setup.py sdist
3、twine upload dist/__packages__-__version__.tar.gz
"""

"""MANIFEST.in
global-include pat1 pat2 ... > Include all files anywhere
global-exclude pat1 pat2 ... > Exclude all files anywhere
graft dir-pattern > Include all files under directories matching dir-pattern
prune dir-pattern > Exclude all files under directories matching dir-pattern

such as: global-exclude *.py[cod]
"""


def read(*parts):
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, *parts), 'r', encoding='utf-8').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='loggus',
    version=find_version('loggus', 'pci.py'),
    description="This is a log library, you can output json and care fields easy.",
    long_description="see https://github.com/CzaOrz/loggus",
    author='CzaOrz',
    author_email='972542644@qq.com',
    url='https://github.com/CzaOrz/loggus',
    packages=find_packages(),
    install_requires=[
        "colorama",
        "coverage",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)
